from pathlib import Path
import numpy as np
from kilosort import io


def mean_waveform(cluster_id, results_dir, n_spikes=np.inf, bfile=None, best=True):
    """Get mean waveform for `n_spikes` random spikes assigned to `cluster_id`.

    Parameters
    ----------
    cluster_id : int
        Cluster index to reference from `spike_clusters.npy` in the results
        directory. Only waveforms from spikes assigned to this cluster will
        be used.
    results_dir : str or Path
        Path to directory where Kilosort4 sorting results were saved.
    n_spikes : int; default=np.inf
        Number of spikes to use for computing mean. By default, all spikes
        assigned to `cluster_id` are used.
    bfile : kilosort.io.BinaryFiltered; optional
        Kilosort4 data file object. By default, this will be loaded using the
        information in `ops.npy` in the saved results.
    best : bool; default=True
        If True, return the mean single-channel waveform using the best channel
        for `cluster_id`. Otherwise, return the multi-channel waveform with
        all data channels included.

    Returns
    -------
    mean_wave : np.ndarray
        Mean waveform with shape (nt,) if `best = True`, or shape (n_chans,nt)
        otherwise.
    
    """
    results_dir = Path(results_dir)
    if best:
        chan = get_best_channels(results_dir)[cluster_id]
    else:
        chan = None

    spikes = get_cluster_spikes(cluster_id, results_dir, n_spikes=n_spikes)
    waves = get_spike_waveforms(spikes, results_dir, chan=chan)
    mean_wave = waves.mean(axis=-1)

    return mean_wave


def get_best_channels(results_dir):
    """Get channel numbers with largest template norm for each cluster."""
    templates = np.load(results_dir / 'templates.npy')
    best_chans = (templates**2).sum(axis=1).argmax(axis=-1)
    return best_chans

def get_best_channel(results_dir, cluster_id):
    return get_best_channels(results_dir)[cluster_id]

def get_cluster_spikes(cluster_id, results_dir, n_spikes=np.inf):
    """Get `n_spikes` random spike times assigned to `cluster_id`."""
    spike_times = np.load(results_dir / 'spike_times.npy')
    spike_clusters = np.load(results_dir / 'spike_clusters.npy')
    spikes = spike_times[spike_clusters == cluster_id]
    if n_spikes != np.inf:
        spikes = np.random.choice(
            spikes, min(spikes.size, n_spikes), replace=False
            )
        spikes.sort()

    return spikes


def get_spike_waveforms(spikes, results_dir, bfile=None, chan=None):
    """Get waveform for each spike in `spikes`, multi- or single-channel.
    
    Parameters
    ----------
    spikes : list or array-like
        Spike times (in units of samples) for the desired waveforms, from
        `spike_times.npy`.
    results_dir : str or Path
        Path to directory where Kilosort4 sorting results were saved.
    bfile : kilosort.io.BinaryFiltered; optional
        Kilosort4 data file object. By default, this will be loaded using the
        information in `ops.npy` in the saved results.
    chan : int; optional.
        Channel to use for single-channel waveforms. If not specified, all
        channels will be returned.

    Returns
    -------
    waves : np.ndarray
        Array of spike waveforms with shape `(nt, len(spikes))`.
    
    """
    if isinstance(spikes, int):
        spikes = [spikes]

    if bfile is None:
        ops = io.load_ops(results_dir / 'ops.npy')
        bfile = io.bfile_from_ops(ops)
    whitening_mat_inv = np.load(results_dir / 'whitening_mat_inv.npy')

    waves = []
    for t in spikes:
        tmin = t - bfile.nt0min
        tmax = t + (bfile.nt - bfile.nt0min)
        w = bfile[tmin:tmax].cpu().numpy()
        if whitening_mat_inv is not None:
            w = whitening_mat_inv @ w
        if w.shape[1] == bfile.nt:
            # Don't include spikes at the start or end of the recording that
            # get truncated to fewer time points.
            waves.append(w)
    waves = np.stack(waves, axis=-1)

    if chan is not None:
        waves = waves[chan,:]
    
    bfile.close()

    return waves


def cluster_templates(cluster_id, results_dir, mean=False, best=True):
    """Get templates assigned to all spikes assigned this `cluster_id.`
    
    Parameters
    ----------
    cluster_id : int
        Cluster index to reference from `spike_clusters.npy` in the results
        directory. Only waveforms from spikes assigned to this cluster will
        be used.
    results_dir : str or Path
        Path to directory where Kilosort4 sorting results were saved.
    mean : bool; default=False
        If True, return the mean across templates.
    best : bool; default=True
        If True, return single channel template(s) for this cluster_id using
        the channel with the largest norm.

    Return
    ------
    mean_temps : np.ndarray
        Array of templates with shape `(n_templates, nt, n_channels)`, or
        with shape `(nt, n_channels)` if `mean=True`.
        
    """
    results_dir = Path(results_dir)
    spike_clusters = np.load(results_dir / 'spike_clusters.npy')
    spike_idx = spike_clusters[spike_clusters == cluster_id]
    temps = get_templates(spike_idx, results_dir)
    if best:
        chan = (temps**2).sum(axis=1).sum(axis=0).argmax(axis=-1)
        temps = temps[:,:,chan]
    if mean:
        return temps.mean(axis=0)

    return temps


def get_templates(spike_idx, results_dir):
    """Get learned templates assigned to one or more spikes.
    
    Parameters
    ----------
    spike_idx : int or array-like
        Index or list/array of indices into `spike_times.npy`
    results_dir : str or Path
        Path to directory where Kilosort4 sorting results were saved.
    
    Returns
    -------
    scaled : np.ndarray
        Array of templates with shape `(len(spike_idx), nt, n_channels)`.
        Templates are scaled using `amplitude.npy` to match the scale of
        unwhitened spike waveforms.

    """
    results_dir = Path(results_dir)
    if isinstance(spike_idx, int):
        spike_idx = [spike_idx]
    templates = np.load(results_dir / 'templates.npy')
    spike_templates = np.load(results_dir / 'spike_templates.npy')
    amplitudes = np.load(results_dir / 'amplitudes.npy')

    template_idx = spike_templates[spike_idx]
    temps = templates[template_idx, :, :]
    scaled = amplitudes[spike_idx, np.newaxis, np.newaxis] * temps

    return scaled


def get_good_cluster(results_dir, n=1):
    """Pick `n` random cluster ids with a label of 'good.'"""
    labels = get_labels(results_dir)
    labels = [x for x in labels if x[1] == 'good']
    rows = np.random.choice(
        np.arange(len(labels)), size=min(n, len(labels)), replace=False
        )
    cluster_ids = [int(labels[r][0]) for r in rows]
    if n == 1:
        cluster_ids = cluster_ids[0]

    return cluster_ids


def get_labels(results_dir):
    """Load good/mua labels as a list of ['cluster', 'label'] pairs."""
    results_dir = Path(results_dir)
    filename = results_dir / 'cluster_KSLabel.tsv'
    with open(filename) as f:
        text = f.read()
    rows = text.split('\n')
    labels = [r.split('\t') for r in rows[1:]][:-1]

    return labels



import numpy as np
from numba import njit
import math
from kilosort.CCG import compute_CCG, CCG_metrics

def count_elements(kk, iclust, my_clus, xtree):
    n1 = np.isin(iclust, my_clus[xtree[kk, 0]]).sum()
    n2 = np.isin(iclust, my_clus[xtree[kk, 1]]).sum()
    return n1, n2

def check_split(Xd, kk, xtree, iclust, my_clus):
    ixy = np.isin(iclust, my_clus[xtree[kk, 2]])
    iclu = iclust[ixy]
    labels = 2*np.isin(iclu, my_clus[xtree[kk, 0]]) - 1

    Xs = Xd[ixy]
    Xs[:,-1] = 1

    w = np.ones((Xs.shape[0],1))
    w[labels>0] = np.mean(labels<0)
    w[labels<0] = np.mean(labels>0)

    CC = Xs.T @ (Xs * w)
    CC = CC + .01 * np.eye(CC.shape[0])
    b = np.linalg.solve(CC, labels @ (Xs * w))
    xproj = Xs @ b

    score = bimod_score(xproj)
    return xproj, score

def clean_tree(valid_merge, xtree, inode):
    ix = (xtree[:,2]==inode).nonzero()[0]
    if len(ix)==0:
        return
    valid_merge[ix] = 0
    clean_tree(valid_merge, xtree, xtree[ix, 0])
    clean_tree(valid_merge, xtree, xtree[ix, 1])
    return

def bimod_score(xproj):
    from scipy.ndimage import gaussian_filter1d
    xbin, _ = np.histogram(xproj, np.linspace(-2,2,400))
    xbin = gaussian_filter1d(xbin.astype('float32'), 4)

    imin = np.argmin(xbin[175:225])
    xmin = np.min(xbin[175:225])
    xm1  = np.max(xbin[:imin+175])
    xm2  = np.max(xbin[imin+175:])

    score = 1 - np.maximum(xmin/xm1, xmin/xm2)
    return score

def check_CCG(st1, st2=None, nbins = 500, tbin  = 1/1000):
    if st2 is None:
        st2 = st1.copy()
    K , T= compute_CCG(st1, st2, nbins = nbins, tbin = tbin)
    Q12, R12, R00 = CCG_metrics(st1, st2, K, T,  nbins = nbins, tbin = tbin)
    is_refractory    = Q12<.1  and (R12<.2  or R00<.25)
    cross_refractory = Q12<.25 and (R12<.05 or R00<.25)
    return is_refractory, cross_refractory

def refractoriness(st1, st2):
    # compute goodness of st1, st2, and both

    is_refractory = check_CCG(st1, st2)[1]
    if is_refractory:
        criterion = 1 # never split
        #print('this is refractory')
    else:
        criterion = 0
        #good_0 = check_CCG(np.hstack((st1,st2)))[0]
        #good_1 = check_CCG(st1)[0]
        #good_2 = check_CCG(st2)[0]
        #print(good_0, good_1, good_2)
        #if (good_0==1) and (good_1==0) and (good_2==0):
        #    criterion = 1 # don't split
        #    print('good cluster becomes bad')
    return criterion

def split(Xd, xtree, tstat, iclust, my_clus, verbose = True, meta = None):
    xtree = np.array(xtree)

    kk = xtree.shape[0]-1
    nc = xtree.shape[0] + 1
    valid_merge = np.ones((nc-1,), 'bool')


    for kk in range(nc-2,-1,-1):
        if not valid_merge[kk]:
            continue;

        ix1 = np.isin(iclust, my_clus[xtree[kk, 0]])
        ix2 = np.isin(iclust, my_clus[xtree[kk, 1]])

        criterion = 0
        score = np.NaN
        if criterion==0:
            # first mutation is global modularity
            if tstat[kk,0] < 0.2:
                criterion = -1


        if meta is not None and criterion==0:
            # second mutation is based on meta_data
            criterion = refractoriness(meta[ix1],meta[ix2])
            #criterion = 0
        
        if criterion==0:
            xproj, score = check_split(Xd, kk, xtree, iclust, my_clus)
            # third mutation is bimodality
            #xproj, score = check_split(Xd, kk, xtree, iclust, my_clus)
            criterion = 2 * (score <  .6) - 1

        if criterion==0:
            # fourth mutation is local modularity (not reachable)
            score = tstat[kk,-1]
            criterion = score > .15

        if verbose:
            n1,n2 = ix1.sum(), ix2.sum()
            #print('%3.0d, %6.0d, %6.0d, %6.0d, %2.2f,%4.2f, %2.2f'%(kk, n1, n2,n1+n2,
            #tstat[kk,0], tstat[kk,-1], score))

        if criterion==1:
            valid_merge[kk] = 0
            clean_tree(valid_merge, xtree, xtree[kk,0])
            clean_tree(valid_merge, xtree, xtree[kk,1])

    tstat = tstat[valid_merge]
    xtree = xtree[valid_merge]

    return xtree, tstat


def new_clusters(iclust, my_clus, xtree, tstat):

    if len(xtree)==0:
        return np.zeros_like(iclust)
         

    nc = xtree.max() + 1

    isleaf = np.zeros(2*nc-1,)
    isleaf[xtree[:,0]] = 1
    isleaf[xtree[:,1]] = 1
    isleaf[xtree[:,2]] = 0

    ind = np.nonzero(isleaf)[0]
    iclust1 = iclust.copy()
    for j in range(len(ind)):
        ix = np.isin(iclust, my_clus[ind[j]])
        iclust1[ix] = j
        xtree[xtree[:,0] == ind[j], 0] = j
        xtree[xtree[:,1] == ind[j], 1] = j


    return iclust1



from numba import njit
import numpy as np
import torch
from torch.nn.functional import conv1d
import math 
from tqdm import trange 

@njit()
def compute_CCG(st1, st2, tbin = 1/1000, nbins = 500):

    st1 = np.sort(st1)
    st2 = np.sort(st2)

    dt = nbins * tbin
    T = np.maximum(st1.max(), st2.max()) - np.minimum(st1.min(), st2.min())

    ilow = 0
    ihigh = 0
    j = 0

    K = np.zeros(2*nbins+1,)
    while j<len(st2):
        while (ihigh<len(st1)) and (st1[ihigh] <= st2[j]+dt):
            ihigh += 1
        while (ilow<len(st1))  and (st1[ilow] <= st2[j]-dt):
            ilow += 1
        if ilow >=len(st1):
            break
        if st1[ilow]>st2[j]+dt:
            j += 1
            continue;
        for k in range(ilow, ihigh):
            ibin = int(np.round((st2[j] - st1[k])/tbin))
            K[ibin+nbins] += 1
        j += 1
    return K, T

#@njit()
def CCG_metrics(st1, st2, K, T, nbins=None, tbin=None):
    irange1 = np.hstack((np.arange(1,nbins//2), np.arange(3*nbins//2, 2*nbins)))
    irange2 = np.arange(nbins-50, nbins-10)
    irange3 = np.arange(nbins+10, nbins+50)

    Q00 = K[irange1].sum() / (len(irange1) * tbin * len(st1) * len(st2) /T)
    Q1 = K[irange2].sum() / (len(irange2) * tbin * len(st1) * len(st2) /T)
    Q2 = K[irange3].sum() / (len(irange3) * tbin * len(st1) * len(st2) /T)
    Q01 = np.maximum(Q1, Q2)

    R00 = np.maximum(K[irange2].mean(), K[irange3].mean())
    R00 = np.maximum(R00, K[irange1].mean())

    a = K[nbins]
    K[nbins] = 0

    Qi = np.zeros(10,)
    Ri = np.zeros(10,)
    for i in range(1,11):
        irange = np.arange(nbins-i, nbins+i+1)
        Qi0 = K[irange].sum() / (2*i*tbin*len(st1)*len(st2)/T)
        Qi[i-1] = Qi0

        n = K[irange].sum()/2
        lam = R00 * i

        p = 1/2 * (1 + math.erf((n-lam)/(1e-10 + 2*lam)**.5))

        Ri[i-1] = p

    K[nbins] = a
    Q12 = np.min(Qi) / (1e-10 + np.maximum(Q00, Q01))
    R12 = np.min(Ri)

    #print('%4.2f, %4.2f, %4.2f'%(R00, Q12, R12))
    return Q12, R12, R00

def check_CCG(st1, st2=None, nbins = 500, tbin  = 1/1000, acg_threshold=0.2,
              ccg_threshold=0.25):
    if st2 is None:
        st2 = st1.copy()
    K , T= compute_CCG(st1, st2, nbins = nbins, tbin = tbin)
    Q12, R12, R00 = CCG_metrics(st1, st2, K, T,  nbins = nbins, tbin = tbin)
    is_refractory    = Q12<acg_threshold  and (R12<.2)#  or R00<.25)
    cross_refractory = Q12<ccg_threshold and (R12<.05)# or R00<.25)
    return is_refractory, cross_refractory, Q12

def similarity(Wall, W, nt=61):
    WtW = conv1d(W.reshape(-1, 1,nt), W.reshape(-1, 1 ,nt), padding = nt) 
    WtW = torch.flip(WtW, [2,])
    mu = (Wall**2).sum((1,2), keepdims=True)**.5
    Wnorm = Wall / (1e-6 + mu)
    UtU = torch.einsum('ilk, jlm -> ijkm',  Wnorm, Wnorm)
    similar_templates = torch.einsum('ijkm, kml -> ijl', UtU.cpu(), WtW.cpu()).numpy()
    similar_templates = similar_templates.max(axis=-1)
    return similar_templates

def refract(iclust2, st0, acg_threshold=0.2, ccg_threshold=0.25):
    
    Nfilt = iclust2.max()+1

    is_refractory    = np.zeros(Nfilt, )
    cross_refractory = np.zeros(Nfilt, )
    Q12 = np.zeros(Nfilt, )

    for kk in range(Nfilt):    
        ix = iclust2==kk
        st1 = st0[ix]

        if (len(st1) > 10) and ((st1.max() - st1.min()) != 0):
            is_refractory[kk], cross_refractory[kk], Q12[kk] = check_CCG(
                st1, acg_threshold=acg_threshold, ccg_threshold=ccg_threshold
                )

    return is_refractory, Q12


import time
import logging
logger = logging.getLogger(__name__)

import numpy as np
import torch
from qtpy import QtCore

import kilosort
from kilosort.run_kilosort import (
    setup_logger, initialize_ops, compute_preprocessing, compute_drift_correction,
    detect_spikes, cluster_spikes, save_sorting, close_logger
    )
from kilosort.io import save_preprocessing
from kilosort.utils import (
    log_performance, log_cuda_details, ops_as_string, probe_as_string
    )


class KiloSortWorker(QtCore.QThread):
    finishedPreprocess = QtCore.Signal(object)
    finishedSpikesort = QtCore.Signal(object)
    finishedAll = QtCore.Signal(object)
    progress_bar = QtCore.Signal(int)
    plotDataReady = QtCore.Signal(str)

    def __init__(self, context, results_directory, steps,
                 device=torch.device('cuda'), file_object=None, *args, **kwargs):
        super(KiloSortWorker, self).__init__(*args, **kwargs)
        self.context = context
        self.data_path = context.data_path
        self.results_directory = results_directory
        assert isinstance(steps, list) or isinstance(steps, str)
        self.steps = steps if isinstance(steps, list) else [steps]
        self.device = device
        self.file_object = file_object

    def run(self):
        if "spikesort" in self.steps:
            settings = self.context.params
            probe = self.context.probe
            settings["data_dir"] = self.data_path.parent
            settings["filename"] = self.data_path
            results_dir = self.results_directory
            if not results_dir.exists():
                results_dir.mkdir(parents=True)
            
            setup_logger(results_dir)
            verbose = settings['verbose_log']

            try:
                logger.info(f"Kilosort version {kilosort.__version__}")
                logger.info(f"Sorting {self.data_path}")
                clear_cache = settings['clear_cache']
                if clear_cache:
                    logger.info('clear_cache=True')
                logger.info('-'*40)

                tic0 = time.time()

                if probe['chanMap'].max() >= settings['n_chan_bin']:
                    raise ValueError(
                        f'Largest value of chanMap exceeds channel count of data, '
                        'make sure chanMap is 0-indexed.'
                    )

                if settings['nt0min'] is None:
                    settings['nt0min'] = int(20 * settings['nt']/61)
                data_dtype = settings['data_dtype']
                device = self.device
                save_preprocessed_copy = settings['save_preprocessed_copy']
                do_CAR = settings['do_CAR']
                invert_sign = settings['invert_sign']
                if not do_CAR:
                    logger.info("Skipping common average reference.")

                ops = initialize_ops(settings, probe, data_dtype, do_CAR,
                                     invert_sign, device, save_preprocessed_copy)

                # Pretty-print ops and probe for log
                logger.debug(f"Initial ops:\n\n{ops_as_string(ops)}\n")
                logger.debug(f"Probe dictionary:\n\n{probe_as_string(ops['probe'])}\n")

                # TODO: add support for file object through data conversion
                # Set preprocessing and drift correction parameters
                ops = compute_preprocessing(ops, self.device, tic0=tic0,
                                            file_object=self.file_object)
                np.random.seed(1)
                torch.cuda.manual_seed_all(1)
                torch.random.manual_seed(1)
                ops, bfile, st0 = compute_drift_correction(
                    ops, self.device, tic0=tic0, progress_bar=self.progress_bar,
                    file_object=self.file_object, clear_cache=clear_cache
                    )

                # Check scale of data for log file
                b1 = bfile.padded_batch_to_torch(0).cpu().numpy()
                logger.debug(f"First batch min, max: {b1.min(), b1.max()}")

                if save_preprocessed_copy:
                    save_preprocessing(results_dir / 'temp_wh.dat', ops, bfile)

                # Will be None if nblocks = 0 (no drift correction)
                if st0 is not None:
                    self.dshift = ops['dshift']
                    self.st0 = st0
                    self.plotDataReady.emit('drift')

                # Sort spikes and save results
                st, tF, Wall0, clu0 = detect_spikes(
                    ops, self.device, bfile, tic0=tic0,
                    progress_bar=self.progress_bar, clear_cache=clear_cache,
                    verbose=verbose
                    )

                self.Wall0 = Wall0
                self.wPCA = torch.clone(ops['wPCA'].cpu()).numpy()
                self.clu0 = clu0
                self.plotDataReady.emit('diagnostics')

                clu, Wall = cluster_spikes(
                    st, tF, ops, self.device, bfile, tic0=tic0,
                    progress_bar=self.progress_bar, clear_cache=clear_cache,
                    verbose=verbose
                    )
                ops, similar_templates, is_ref, est_contam_rate, kept_spikes = \
                    save_sorting(ops, results_dir, st, clu, tF, Wall, bfile.imin, tic0)

            except Exception as e:
                if isinstance(e, torch.cuda.OutOfMemoryError):
                    logger.exception('Out of memory error, printing performance...')
                    log_performance(logger, level='info')
                    log_cuda_details(logger)
                # This makes sure the full traceback is written to log file.
                logger.exception('Encountered error in `run_kilosort`:')
                # Annoyingly, this will print the error message twice for console
                # but I haven't found a good way around that.
                raise

            finally:
                close_logger()

            self.ops = ops
            self.st = st[kept_spikes]
            self.clu = clu[kept_spikes]
            self.tF = tF[kept_spikes]
            self.is_refractory = is_ref
            self.plotDataReady.emit('probe')

            self.finishedSpikesort.emit(self.context)



import os, tempfile, shutil, pathlib, psutil
import importlib.util
import logging
import pprint
logger = logging.getLogger(__name__)

import numpy as np
from tqdm import tqdm
from urllib.request import urlopen
from urllib.error import HTTPError

import torch

# pynvml is an optional dependency for Pytorch cuda, so existing installations
# may or may not have it. It's not worth causing an interruption for this.
if importlib.util.find_spec('pynvml') is not None:
    _NVML_EXISTS = True
else:
    _NVML_EXISTS = False

_DOWNLOADS_URL = 'https://www.kilosort.org/downloads'
_DOWNLOADS_DIR_ENV = os.environ.get("KILOSORT_LOCAL_DOWNLOADS_PATH")
_DOWNLOADS_DIR_DEFAULT = pathlib.Path.home().joinpath('.kilosort')
DOWNLOADS_DIR = pathlib.Path(_DOWNLOADS_DIR_ENV) if _DOWNLOADS_DIR_ENV else _DOWNLOADS_DIR_DEFAULT
PROBE_DIR = DOWNLOADS_DIR.joinpath('probes')

# use mat file probes because they enable disconnected channels
probe_names = [
    'neuropixPhase3A_kilosortChanMap.mat',
    'neuropixPhase3B1_kilosortChanMap.mat',\
    'neuropixPhase3B2_kilosortChanMap.mat',
    'NP2_kilosortChanMap.mat', 
    'Linear16x1_kilosortChanMap.mat',
    ]

def template_path(basename='wTEMP.npz'):
    """ currently only one set of example templates to use"""
    return cache_template_path(basename)

def cache_template_path(basename):
    DOWNLOADS_DIR.mkdir(parents=True, exist_ok=True)
    url = f'{_DOWNLOADS_URL}/{basename}'
    cached_file = os.fspath(DOWNLOADS_DIR.joinpath(basename)) 
    if not os.path.exists(cached_file):
        logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
        download_url_to_file(url, cached_file, progress=True)
    return cached_file

def download_probes(probe_dir=None):
    if probe_dir is None:
        probe_dir = PROBE_DIR
    probe_dir.mkdir(parents=True, exist_ok=True)
    for probe_name in probe_names:
        url = f'{_DOWNLOADS_URL}/{probe_name}'
        cached_file = os.fspath(probe_dir.joinpath(probe_name)) 
        if not os.path.exists(cached_file):
            logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
            try:
                download_url_to_file(url, cached_file, progress=True)
            except HTTPError as e:
                logger.info(f'Unable to download probe {probe_name}, error:')
                logger.info(e)


def download_url_to_file(url, dst, progress=True):
    r"""Download object at the given URL to a local path.
            Thanks to torch, slightly modified
    Args:
        url (string): URL of the object to download
        dst (string): Full path where object will be saved, e.g. `/tmp/temporary_file`
        progress (bool, optional): whether or not to display a progress bar to stderr
            Default: True
    """
    file_size = None
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context
    u = urlopen(url)
    meta = u.info()
    if hasattr(meta, 'getheaders'):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])
    # We deliberately save it in a temp file and move it after
    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)
    try:
        with tqdm(total=file_size, disable=not progress,
                  unit='B', unit_scale=True, unit_divisor=1024) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                pbar.update(len(buffer))
        f.close()
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def log_performance(log=None, level=None, header=None):
    """Log usage information for cpu, memory, gpu, and gpu memory.

    Parameters
    ----------
    log : logging.Logger; optional.
        Logger object used to write the text. If not provided, the logger for
        `kilosort.utils` will be used.
    level : str; optional.
        Logging level to use. By default, 'debug' will be used. See documentation
        for the built-in `logging` module for information on logging levels.
    header : str; optional.
        Text to output before usage information. For example, an iteration number
        when logging within a loop.

    Notes
    -----
    Usage values include resources used by non-Kilosort processes. This is done
    so that, in the event of an error or crash, we can see if it was due to
    exceeding system resources regardless of whether Kilosort itself was using
    all of them.

    """

    if log is None:
        log = logger
    if level is None:
        level = 'debug'
    if header is not None:
        getattr(log, level)(' ')
        getattr(log, level)(f'{header}')

    getattr(log, level)('*'*56)
    # TODO: This part is slow, on the order of ms versus micro-seconds for GPU check.
    #       Find a faster way to check main memory and cpu usage.
    getattr(log, level)(f'CPU usage:    {psutil.cpu_percent():5.2f} %')

    memory = psutil.virtual_memory()
    used = memory.used / 2**30
    total = memory.total / 2**30
    pct = (used / total) * 100
    getattr(log, level)(f'Memory:       {pct:5.2f} %     |{used:10.2f}   / {total:8.2f} GB')
    getattr(log, level)('-'*54)

    if torch.cuda.is_available():
        if _NVML_EXISTS:
            getattr(log, level)(f'GPU usage:    {torch.cuda.utilization():5.2f} %')
        else:
            getattr(log, level)(f'GPU usage:    `conda install pynvml` for GPU usage')

        gpu_avail, gpu_total = torch.cuda.mem_get_info()
        gpu_avail /= (2**30)  # convert bytes -> gb
        gpu_total /= (2**30)
        gpu_used = gpu_total - gpu_avail
        gpu_pct = (gpu_used / gpu_total) * 100

        getattr(log, level)(f'GPU memory:   {gpu_pct:5.2f} %     |{gpu_used:10.2f}   / {gpu_total:8.2f} GB')
        allocated = torch.cuda.memory_allocated() / 2**30
        alloc_pct = (allocated / gpu_total) * 100
        getattr(log, level)(f'Allocated:    {alloc_pct:5.2f} %     |{allocated:10.2f}   / {gpu_total:8.2f} GB')
        max_alloc = torch.cuda.max_memory_allocated() / 2**30
        max_pct = (max_alloc / gpu_total) * 100
        getattr(log, level)(f'Max alloc:    {max_pct:5.2f} %     |{max_alloc:10.2f}   / {gpu_total:8.2f} GB')
    else:
        getattr(log, level)('GPU usage:    N/A')
        getattr(log, level)('GPU memory:   N/A')

    getattr(log, level)('*'*56)


def log_cuda_details(log=None):
    """Log a detailed summary of cuda stats from `torch.cuda.memory_summary`."""
    if log is None: log = logger
    if torch.cuda.is_available():
        log.debug(f'\n\n{torch.cuda.memory_summary(abbreviated=True)}\n')


def probe_as_string(probe):
    """Format probe dictionary as copy-pasteable-to-code string."""

    # Set numpy to print full arrays
    opt = np.get_printoptions()
    np.set_printoptions(threshold=np.inf)
    
    p = pprint.pformat(probe, indent=4, sort_dicts=False)
    # insert `np.` so that text can be copied directly to code
    p = 'np.array'.join(p.split('array'))
    p = 'dtype=np.'.join(p.split('dtype='))
    probe_text = "probe = "
    # Put curly braces on separate lines
    probe_text += p[0] + '\n ' + p[1:-1] + '\n' + p[-1]

    # Revert numpy settings
    np.set_printoptions(**opt)

    return probe_text


def ops_as_string(ops):
    """Format ops dictionary as copy-pasteable-to-code string.
    
    Notes
    -----
    Keys for `settings` and `probe` are removed since they contain a lot of
    redundant information and are difficult to format in a nested way. See
    `probe_as_string` for printing probe information.
    
    """

    ops_copy = ops.copy()
    probe_keys = list(ops['probe'].keys())
    for k in ['settings', 'probe'] + probe_keys:
        _ = ops_copy.pop(k)
    ops_text = "ops = "
    p = pprint.pformat(ops_copy, indent=4, sort_dicts=False)
    # Put curly braces on separate lines
    ops_text += p[0] + '\n ' + p[1:-1] + '\n' + p[-1]

    return ops_text



import setuptools
from setuptools import setup

install_deps = [
    'numpy>=1.20.0,<2.0.0',
    'scipy',
    'scikit-learn',
    'tqdm',
    'torch>=1.6',
    'numba',
    'faiss-cpu',
    'psutil'
]

gui_deps = [
    'pyqtgraph>=0.13.0',
    'qtpy',
    'pyqt6',
    'pyqt6.sip',
    'matplotlib'
]

docs_deps = [
    'sphinx>=3.0',
    'sphinxcontrib-apidoc',
    'nbsphinx',
    'myst_parser',
    'sphinx_rtd_theme',
    'pandoc'
]

### remove torch install if already installed
try:
    import torch
    a = torch.ones(2, 3)
    version = int(torch.__version__.split(".")[1])
    if version >= 6:
        install_deps.remove("torch>=1.6")
except:
    pass

### remove pyqt6 install if other qt backend installed
try:
    import PyQt5
    gui_deps.remove("pyqt6")
    gui_deps.remove("pyqt6.sip")
    gui_deps.append("pyqt5")
    gui_deps.append("pyqt5.sip")
except:
    pass

try:
    import PySide2
    gui_deps.remove("pyqt6")
    gui_deps.remove("pyqt6.sip")
except:
    pass

try:
    import PySide6
    gui_deps.remove("pyqt6")
    gui_deps.remove("pyqt6.sip")
except:
    pass

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="kilosort",
    license="BSD",
    author="Marius Pachitariu",
    author_email="pachitarium@janelia.hhmi.org",
    description="spike sorting pipeline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MouseLand/kilosort",
    setup_requires=[
        'pytest-runner',
        'setuptools-scm',
    ],
    packages=setuptools.find_packages(),
    use_scm_version=True,
    install_requires=install_deps,
    tests_require=[
        'pytest'
    ],
    extras_require={
        'docs': docs_deps,
        'gui': gui_deps,
        'all': gui_deps,
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
        ]
    }
)
